//*****************************************************************************/
//
// Filename: WireTapClientAPI.h
//
// Copyright (c) 2017 Autodesk, Inc.
// All rights reserved.
// 
// Use of this software is subject to the terms of the Autodesk license
// agreement provided at the time of installation or download, or which
// otherwise accompanies this software in either electronic or hard copy form.
//*****************************************************************************/

#ifndef DL_WIRETAPCLIENTAPI_H
#define DL_WIRETAPCLIENTAPI_H

//=============================================================================
// I N C L U D E   F I L E S

#ifndef DL_WIRETAPTYPES_H
#include <WireTapTypes.h>
#endif

//=============================================================================
// F O R W A R D   D E C L A R A T I O N S

class WireTapNodeImp;
class WireTapServerImp;
class WireTapNodeHandle;
class WireTapPathTranslationGroup;
class WireTapHostTranslation;
class WireTapOsTypeTranslation;

#ifndef SKIP_DOXYGEN_MAIN_PAGE_DEFINITION
/// @mainpage Autodesk Wiretap Client C++ API Reference
///
/// <h6 align="center"> 2003-2016 Autodesk Canada Co./Autodesk Inc.</h6> 
/// <H3> Some Pointers for Using this Wiretap Documentation </H3>
///
/// <H5>To view a class and its members:</H5>
/// <ol>
///   <li>Click Class List. 
///   <li>Click the name of the class you want to view.
/// </ol>
///
/// <H5>To find a function by name:</H5>
/// <ol>
///   <li>Click Class Members. 
///   <li>Click the appropriate alphabetical tab (labelled b to ~). 
///   <li>Click the function you want to view. 
/// </ol>
///
#endif // #ifndef SKIP_DOXYGEN_MAIN_PAGE_DEFINITION

//=============================================================================
// P U B L I C   F U N C T I O N S

/// @brief  Initializes the Wiretap Client API. 
///
/// This function must be called once at startup, 
/// before any other calls to the Wiretap Client API are made.
///
/// @return  true if the Client API was initialized successfully; false if the 
/// API was not initialized. On success, the rest of the Wiretap Client API can 
/// be used. 
///
/// @see WireTapClientUninit
///
WIRETAP_EXPORT_SPECIFIER bool WireTapClientInit();

/// @brief  Release unused shared server connections.
///
/// It will release shared connections that have been unused for more than
/// 10 seconds, that is, that are no longer tied to any WireTapServerHandle.
///
/// @param  force  An input parameter. If true, it will release all unused
///                connections independently from how long they have been unused.
///
/// @see WireTapClientInit
///
WIRETAP_EXPORT_SPECIFIER void WireTapClientReleaseUnusedConnections(bool force);

/// @brief  Register a callback function for server disconnections.
///
/// This function asks that the provided callback function be called when 
/// a server forced our disconnection (due to a crash or an externally  
/// requested shutdown). It will not be called following a client-initiated 
/// disconnection.
///
/// @note The callback function will be called only from within a Wiretap
///       API call. 
///
/// This callback replaces the need for calling isConnected() after every 
/// reported error. It can then be used to flag the need for server
/// reinitialization.  (Remember that once the server is restarted, following
/// Wiretap API calls will automatically reconnect to the server.)
///
/// @note Do not make Wiretap API calls in the callback function as this 
///       may cause a dead-lock. The typical procedure is to set a flag which
///       will be checked before calling the Wiretap API again. 
///
/// @note Callback is registered until replaced by another function or NULL.
///       Only one callback can be registered. 
///
/// @param  cb  A callback function which takes as argument the serverId 
///             initially passed to the WireTapServerHandle constructor.
///
WIRETAP_EXPORT_SPECIFIER void WireTapRegisterDisconnectCallBack(
                                    void cb(const WireTapServerId&) );

/// @brief  Uninitializes the Wiretap Client API. 
///
/// This function must be called at shutdown.
///
/// @warning No further Wiretap Client SDK function calls should be done
///          after calling this function, including calls done by static
///          objects if this is called from main().
///
/// @see WireTapClientInit
///
WIRETAP_EXPORT_SPECIFIER void WireTapClientUninit();

/// @brief  Scope base Wiretap client
///
/// This object will initialize the Wiretap Client API when init() is called
/// and will unitialize it when the object falls out of scope.
///
class WIRETAP_EXPORT_SPECIFIER WireTapClient
{
  bool _initialized;

  public:
    WireTapClient();

    /// @see WireTapClientUninit
    ///
    ~WireTapClient();

    /// @see WireTapClientInit
    ///
    bool init();
};

/// @brief  Gets the version of the Wiretap Client API. 
///
/// Gets the full version of the current instance of the Wiretap 
/// Client API. Useful if it is necessary to display or verify the version of 
/// the API at runtime. The version is obtained in three parts: 
/// major.minor.maintenance (for example: 2008.1.0).
///
/// @param  majorVersion   An output parameter. The function sets this 
///                        parameter to the major version of the API.  
///
/// @param  minorVersion   An output parameter. The function sets this 
///                        parameter to the minor version of the API.
///
/// @param  maintVersion   An output parameter. The function sets this 
///                        parameter to a maintenance suffix. 
///                        If the version is not a maintenance release, this is
///                        is set to 0.
///
WIRETAP_EXPORT_SPECIFIER void WireTapClientGetVersion( int &majorVersion,
                                                       int &minorVersion,
                                                       int &maintVersion );

/// @brief This class represents a connection to a Wiretap server on the Wiretap
/// network.  
///
/// In this context, "Wiretap server" means an instance of a particular Wiretap 
/// server running on a particular host. For example: an instance of the IFFFS 
/// Wiretap Server running on a host machine named "belgium". 
/// 
/// WireTapServerHandle coordinates all exchanges between a Wiretap
/// client and a Wiretap server. It maintains the network connection between the
/// client and the server as long as required &mdash; for example, for the 
/// duration of a user work session. 
/// 
/// To instantiate a WireTapServerHandle, Wiretap clients need a
/// WireTapServerId object that contains the persistent identifier 
/// of the Wiretap server. 
///
/// Wiretap clients can call WireTapServerHandle::getRootNode to obtain 
/// a handle on the root node of the WireTapServer. The WireTapNodeHandle for 
/// the root node can be used to access the rest of the nodes in the hierarchy.
/// 
/// All the member functions of the WireTapServerHandle class return a 
/// boolean value which is set to true on success or false on failure. If the 
/// return value is false, the reason for the failure can be obtained by calling
/// WireTapServerHandle::lastError() which returns the error string.  The string
/// must be used immediately, or stored since it will be overwritten during the 
/// next call to any member function.
///
/// @see WireTapServerId, WireTapServerInfo, WireTapServerList, and the sample 
/// program listAllServers.cpp (which is included in the SDK). 
///
class WIRETAP_EXPORT_SPECIFIER WireTapServerHandle
{ 
public:
    /// @brief Constructs a WireTapServerHandle object using an instance of 
    /// WireTapServerId.
    ///
    /// Wiretap clients can construct a WireTapServerId if the required 
    /// information is known, or they can obtain it from WireTapServerList.
    ///
    /// @param serverId  An input parameter. A WireTapServerId object that 
    ///                  contains the persistent identifier of the Wiretap 
    ///                  server.  
    /// 
    WireTapServerHandle( const WireTapServerId & serverId ); 

    /// @brief Copy constructor. 
    ///
    /// Constructs a new WireTapServerHandle object that duplicates the content 
    /// of the original. 
    ///
    /// @param  serverHandle  An input parameter. The WiretapServerHandle object
    ///                       that is to be copied.
    /// 
    WireTapServerHandle( const WireTapServerHandle &serverHandle );

    /// @brief Destructor.
    ///
    virtual ~WireTapServerHandle();

    /// @brief Gets the root node of the Wiretap server. 
    ///
    /// @param  root  An output parameter. The method fills this parameter
    ///               with an instance of WiretapNodeHandle that gives access 
    ///               to the root node of the Wiretap server.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getRootNode( WireTapNodeHandle &root ) const;

    /// @brief Gets the vendor of the Wiretap server. 
    ///
    /// @param  vendor  An output parameter. The method fills this parameter
    ///                 with an instance of WireTapStr that contains the name of
    ///                 the company that sells the Wiretap server.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getVendor ( WireTapStr &vendor ) const;

    /// @brief Gets the name of the Wiretap server product. 
    ///
    /// @param  product  An output parameter. The method fills this parameter
    ///                  with a WireTapStr object that contains the name of the 
    ///                  the Wiretap server product.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getProduct( WireTapStr &product ) const;

    /// @brief Gets the version of the Wiretap server. 
    ///
    /// @param  major  An output parameter. The method sets this parameter
    ///                to the major version number of the Wiretap server.
    /// 
    /// @param  minor  An output parameter. The method sets this parameter
    ///                to the minor version number of the Wiretap server.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getVersion( int &major, int &minor ) const;
    
    /// @brief Utility method to verify server version.
    ///
    /// @param  greaterOrEqual An output parameter. The method sets this
    ///                        boolean to indicate if the Wiretap server's 
    ///                        version is not older than the version specified
    ///                        in following args.
    /// @param  major, minor, 
    ///         maint, build   Input parameters. The version to compare against.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    ///
    bool compareVersion( bool& greaterOrEqual,
                         int major, int minor=0,
                         int maint=0, int build=0) const;

    /// @brief Gets an instance of WireTapServerInfo that contains the latest
    /// information about the Wiretap server. 
    ///
    /// @param  info   An output parameter. The method fills this parameter
    ///                with an instance of WireTapServerInfo that describes
    ///                the Wiretap server.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getInfo( WireTapServerInfo &info ) const;

    /// @brief Gets an instance of WireTapStr that contains a persistent 
    /// identifier for the storage device currently connected to the 
    /// Wiretap server. 
    ///
    /// @param  storageId   An output parameter. The method fills this parameter
    ///                     with an instance of WireTapStr that contains the
    ///                     persistent identifier of the storage device
    ///                     connected to the Wiretap server. This storageId 
    ///                     is guaranteed to be unique across all Wiretap 
    ///                     servers from all vendors.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getStorageId( WireTapStr &storageId ) const;

    /// @brief Checks if this server hosts the specified database. 
    ///
    /// @param  db   Database string as would be returned from
    ///              WireTapServerInfo::getDatabase() 
    /// 
    /// @return  true on exact match, false otherwise or if no connection. 
    /// 
    bool isDatabase( const char *db ) const;

    /// @brief Gets the version of the Wiretap protocol in use on the host
    /// machine on which the Wiretap server is running. 
    ///
    /// The Wiretap protocol is a network protocol used to communicate between 
    /// Wiretap servers and Wiretap clients.
    /// 
    /// @param  major  An output parameter. The method sets this parameter
    ///                to the major version number of the Wiretap protocol.
    /// 
    /// @param  minor  An output parameter. The method sets this parameter
    ///                to the minor version number of the Wiretap protocol.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool getProtocolVersion( int &major, int &minor ) const;

    /// @brief Pings the Wiretap server. 
    ///
    /// @param  timeoutMS   An input parameter. Timeout in milliseconds.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool ping( unsigned timeoutMS = 30000 ) const;

    /// @brief Translates a set of file paths appropropriate for one host and 
    /// operating system, into paths appropriate for a second host with a 
    /// different or unknown operating system. 
    ///
    /// Wiretap servers can translate network paths to account for mount-point 
    /// differences between hosts and operating systems. For example, a path 
    /// mounted on a unix system as "/CXFS/clip/f.dpx" may be mounted as 
    /// "N:\myclips\f.dpx" on an NT system.
    ///
    /// @see WireTapServerhandle.translatePath()
    ///
    /// @note The input parameters srcHost and dstHost can be set as follows:
    /// <ul>
    ///   <li>The name of the local host, that is, the machine on which the 
    ///       Wiretap client is running (to get it, call WireTapOS.getHostName) 
    ///   <li>The name of the machine on which the Wiretap server is running, 
    ///       (to get it, call WireTapServerHandle.getDisplayName)
    /// </ul>
    /// @param  srcPaths  An input parameter. A WireTapStrList object. 
    ///
    /// @param  dstPaths  An output parameter. A WireTapStrList object that 
    ///                   points to an array of destination paths.
    ///                   The source and destination paths may be the 
    ///                   same object.
    ///
    /// @param  srcHost   An input parameter. The name or IP address of the host
    ///                   machine on which the source paths are located. 
    ///                   By default, this is set to the Wiretap server host 
    ///                   (the value that would be returned by
    ///                   WireTapServerHandle.getDisplayName).
    ///
    /// @param  dstHost   An input parameter. The name or IP address of the host
    ///                   machine on which the paths are to be accessed. 
    ///                   By default, this is set to the local hostname.
    ///
    /// @param  srcOS     An input/output parameter. A constant that  
    ///                   describes the operating system of the source machine. 
    ///                   The constants are defined in WireTapOS::OsType. 
    ///                   By default, this is set to WireTapOS::OS_UNKNOWN.
    ///
    /// @param  dstOS     An input/output parameter. A constant that describes 
    ///                   the operating system of the destination machine. 
    ///                   The constants are defined in WireTapOS::OsType. 
    ///                   By default, this is set to WireTapOS::OS_UNKNOWN.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool translatePaths( const WireTapStrList &srcPaths,
			 WireTapStrList &dstPaths,
                         const char *srcHost = 0,
                         const char *dstHost = 0,
                         WireTapOS::OsType srcOS = WireTapOS::OS_UNKNOWN,
                         WireTapOS::OsType dstOS = WireTapOS::OS_UNKNOWN );

    /// @brief Translates a file path appropropriate for one host and 
    /// operating system, into a path appropriate for a second host with a 
    /// different or unknown operating system. 
    ///
    /// Wiretap servers can translate network paths to account for mount-point 
    /// differences between hosts and operating systems. For example, a path 
    /// mounted on a UNIX system as "/CXFS/clip/f.dpx" may be mounted as 
    /// "N:\myclips\f.dpx" on an NT system.
    ///
    /// @see WireTapServerhandle.translatePaths()
    ///
    /// @param  srcPath  An input parameter. A WireTapStr object containing the
    ///                  source path to be translated. 
    ///
    /// @param  dstPath  An input/output parameter. A WireTapStr object 
    ///                  containing the destination path, the result of the 
    ///                  translation. 
    ///                  The source and destination path may point to the same
    ///                  object.  
    ///
    /// @param  srcHost  An input parameter. The name or IP address of the host
    ///                  machine on which the source paths are located.
    ///                  By default, this is set to the Wiretap server host 
    ///                  (the value that would be returned by
    ///                  WireTapServerHandle.getDisplayName).
    ///
    /// @param  dstHost  An output parameter. The name or IP address of the host
    ///                  machine on which the destination paths are located.
    ///
    /// @param  srcOS    An input/output parameter. A constant that  
    ///                  describes the operating system of the source machine. 
    ///                  The constants are defined in WireTapOS::OsType. 
    ///
    /// @param  dstOS    An input/output parameter. A constant that describes 
    ///                  the operating system of the destination machine. 
    ///                  The constants are defined in WireTapOS::OsType. 
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool translatePath( const WireTapStr &srcPath, WireTapStr &dstPath,
                        const char *srcHost = 0,
                        const char *dstHost = 0,
                        WireTapOS::OsType srcOS = WireTapOS::OS_UNKNOWN,
                        WireTapOS::OsType dstOS = WireTapOS::OS_UNKNOWN );

    /// @brief Reads a frame from the Wiretap server and writes it to a buffer; 
    /// should only be used if the frame ID is known but is not associated with
    /// a node handle.
    ///
    /// Before calling this method, a Wiretap client should know the clip format
    /// that is appropriate for the frame. The buffer size can be obtained by 
    /// calling WireTapClipFormat.frameBufferSize on an instance of 
    /// WireTapClipFormat of the right type (for example: RGB, YUV, and so on).
    ///
    /// @note Use of this method can result in poor performance. 
    ///
    /// @param  frameId     An input parameter. The ID of the frame to be read.
    ///                     If the frameId is NULL or is an empty string, the
    ///                     read is skipped but the read-ahead is still
    ///                     performed.
    ///
    /// @param  buffer      An input parameter. The buffer to which the frame 
    ///                     will be written. This value cannot be NULL.
    ///
    /// @param  bufferSize  An input parameter. The number of bytes in the  
    ///                     the buffer to be used for the frame. 
    /// 
    /// @param  nextFrameId1 An input parameter. The next frame that will likely
    ///                     be read from the server.  This parameter is a hint
    ///                     that allows the server to optimize performance.
    ///                     Set to NULL or the empty string to skip read-ahead.
    ///                     Up to 2 frame ID hints may be provided. 
    /// 
    /// @param  nextFrameId2 @see nextFrameId1
    ///
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool readFrame( const char *frameId, void *buffer, size_t bufferSize,
                    const char *nextFrameId1=0, const char *nextFrameId2=0 ); 

    bool readFrame( const char *frameId,
                    const void *params, size_t paramsSize,
                    void *buffer, size_t bufferSize,
                    const char *nextFrameId1=0, const char *nextFrameId2=0 ); 

    /// @brief  Reads a frame from a buffer and writes it to the Wiretap server.
    ///
    /// Before calling this method, a Wiretap client should know the clip format
    /// that is appropriate for the frame. The buffer size can be obtained by 
    /// calling WireTapClipFormat.frameBufferSize on an instance of 
    /// WireTapClipFormat of the right type (for example: RGB, YUV, and so on).
    ///
    /// @note Use of this method can result in poor performance. 
    ///
    /// @param  frameId     An input parameter. The ID of the frame to be 
    ///                     written.
    ///
    /// @param  buffer      An input parameter. The buffer that contains
    ///                     the frame data. This value cannot be NULL.
    ///
    /// @param  bufferSize  An input parameter. The number of bytes in the  
    ///                     the buffer. 
    ///
    /// @note The sample program createClip.cpp shows how to instantiate a 
    /// WireTapClipFormat and obtain the buffer size from it. 
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool writeFrame( const char *frameId, const void *buffer, size_t bufferSize );

    /// @brief  Reads chunks of a raw data stream from the server to a buffer.
    /// 
    /// A stream is essentially raw data tagged with an identifier and stored
    /// under a node. A stream may be ASCII or binary. 
    ///
    /// When calling readStream, it is not necessary to know in advance how 
    /// many chunks of data are in the stream. A chunk is comprised of an array 
    /// of items. Each call to readStream on a given chunk will return the 
    /// number of items actually read, which should only differ from the 
    /// specified number at the end of the stream.
    ///
    /// Random access is supported. The parameter itemOffset determines which 
    /// chunk of data to read. itemOffset is expressed in items. It is not a 
    /// byte offset.      
    ///
    /// A stream's identifier is normally exposed by the server 
    /// via the node's metadata and can be obtained by calling getStreamId
    /// on the WireTapNodeHandle object for the node. 
    ///
    /// The node metadata normally also includes details about the 
    /// stream, such as its length and format. Node metadata can be accessed
    /// by calling these methods on the node handle object:
    /// <ul>
    ///   <li>WireTapNodeHandle.getMetaData
    ///   <li>WireTapNodeHandle.getNumAvailableMetaDataStreams
    ///   <li>WireTapNodeHandle.getAvailableMetaDataStream
    /// </ul>
    ///
    /// @param streamId     An input parameter. The ID of the stream to be 
    ///                     read. Can be obtained by calling getStreamId
    ///                     on the WireTapNodeHandle object for the node with
    ///                     which the stream is associated.
    ///
    /// @param buffer       An input parameter. The buffer to be used
    ///                     for reading the stream. This value cannot be NULL.
    ///
    /// @param bufferSize   An input parameter. The number of bytes in the  
    ///                     the buffer to be used for reading the stream. 
    ///                     For performance purposes when carrying out a bulk 
    ///                     transfer, bufferSize should be set to at least
    ///                     256KB. Reading a large amount of data a few bytes at
    ///                     a time is extremely taxing on the server.
    /// 
    /// @param itemOffset   An input parameter. The number of the item at which
    ///                     to start reading the stream.
    /// 
    /// @param itemSizeInBytes   An input parameter. The number of bytes in each
    ///                          item in the stream. This information should be
    ///                          in the metadata for the stream.
    ///
    /// @param numItems     An input/output parameter. The number of items to be read
    ///                     from the stream / the number of items actually read.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool readStream( const char *streamId, void *buffer, size_t bufferSize,
                     size_t itemOffset, size_t itemSizeInBytes, size_t &numItems ); 

    /// @brief  Reads a local file and pushes the data to the server as a
    ///         stream.
    /// 
    /// This method truncates the destination stream. The sample program 
    /// copyProject.cpp shows how to use this method as well as 
    /// WireTapServerHandle.pullStream.
    ///
    /// @see WireTapServerHandle.pullStream
    ///
    /// @param  streamId       An input parameter. The ID of the stream to be
    ///                        sent to the server. Can be obtained by calling 
    ///                        getStreamId on the WireTapNodeHandle object for 
    ///                        the node with which the stream is associated.
    ///
    /// @param  localFilePath  An input parameter. The path and name of the
    ///                        local file to be read and converted to a stream.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool pushStream( const char *streamId, const char *localFilePath );

    /// @brief  Pulls data from the server and writes it to a local file.
    /// 
    /// This method truncates the local file. The sample program copyProject.cpp
    /// shows how to use this method as well as WireTapServerHandle.pushStream.
    ///
    /// @see WireTapServerHandle.pushStream
    ///
    /// @param  streamId       An input parameter. The ID of the stream to be 
    ///                        read from the server. Can be obtained by calling 
    ///                        getStreamId on the WireTapNodeHandle object for 
    ///                        the node with which the stream is associated.
    ///
    /// @param  localFilePath  An input parameter. The path and name of the
    ///                        local file to be written.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    /// 
    bool pullStream( const char *streamId, const char *localFilePath );

    /// @brief  Invalidate the server cache content.
    ///
    /// @return  true on success, false on failure. On failure,
    /// WireTapServerHandle::lastError() can be called to get the error string.
    /// 
    bool invalidateCache();

    /// @brief Gets the name of the host machine on which the server is running.
    ///
    /// @note This method is preferred over WireTapServerInfo.getHostname.
    /// 
    /// @return  A string containing the name of the host machine on which the 
    /// server is running, or NULL on error.
    /// lastError() can be called to get the error string.
    /// 
    const char *getHostName() const;

    /// @brief Gets the UUID of the host machine on which the server is running.
    ///
    /// @return  The UUID of the host machine on which the server is running,
    ///          or NULL on error. lastError() can be called to get the error.
    ///
    const char *getHostUUID() const;

    /// @brief Gets the display name of the Wiretap server.
    /// 
    /// @param  name  An output parameter. The method fills this parameter with 
    ///               the display name of the Wiretap server. It may be the same 
    ///               as the name of the host machine on which the server is 
    ///               running.
    /// 
    /// @see WireTapServerHandle.getHostName.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapServerHandle::lastError() can be called to get the error string. 
    ///  
    bool getDisplayName( WireTapStr& name ) const;

    /// @return  The WireTapServerId object that was used to create this
    /// WireTapServerHandle. For complete server information, use getInfo().
    ///
    const WireTapServerId &getId() const;
    
    /// @brief Gets a string that is an identifier for the Wiretap server.
    ///
    /// This is a convenience method for obtaining the server ID string.
    /// Equivalent to calling WireTapServerId.getId.
    ///
    /// @return  A string that is an identifier for the Wiretap server. 
    ///
    /// @note As IP addresses can change for a given server (e.g. on a DHCP
    /// network), it can be preferable to use the persistent storage id (see 
    /// WireTapServerHandle.getStorageId()).
    /// 
    const char *getIdStr() const;
   
    /// @brief  Halts the current Wiretap operation and any pending operations 
    /// on the current node.
    ///
    /// This method is useful when the current operation is taking too long and 
    /// the user wishes to abort the operation.  This method is designed to be 
    /// called from a separate thread and requires that the WireTapNodeHandle 
    /// still exists.
    ///
    /// @param  reason    This optional parameter can be to specify the 'error'
    ///                   to be set on the server handle.
    ///
    void stop( const char * reason = 0 ) const;
    
    /// @brief  Force immediate disconnection with server. Calling this
    /// method is optional (disconnections are eventually performed anyway when 
    /// WireTapServerHandles are destroyed). However, if client app determines
    /// there is a problem with a connection, it may call this method to ensure
    /// a fresh new connection with server is established for following queries.
    ///
    void disconnect();

    /// @brief  Determines if there is an active connection to the destination
    /// server.  Connections are established automatically on demand when
    /// calling methods on WireTapServerhandle or WireTapNodehandle objects.
    ///
    /// This method is useful when the client wants to know if the server is
    /// still up (and has not been restarted) without having to ping the server.
    ///
    bool isConnected() const;

    /// @brief Gets the error string for the last operation that failed.  
    ///
    /// Wiretap clients can obtain this error message for any member function
    /// that returns a boolean. The strings can be localized if required. 
    /// See the Wiretap Developer's Guide. 
    ///
    /// @return The error string for the last operation (called on the current 
    /// server handle) that failed.
    ///
    const char *lastError() const;

    #ifndef WIRETAP_STRIP_FROM_DOXYGEN
    WireTapServerImp &imp() const;
    #endif // WIRETAP_STRIP_FROM_DOXYGEN

    // This class is used to retrieve information of a group defined in the
    // path translation database.
    //
    class WIRETAP_EXPORT_SPECIFIER TranslationGroup
    {
    public:
        TranslationGroup();
        ~TranslationGroup();
        TranslationGroup & operator=( const WireTapPathTranslationGroup & i);

        const char *getGroupName() const;
        const char *getOSType() const;
        const char *getHost(int index) const;
        int  getNumHosts() const;

    private:
        WireTapPathTranslationGroup *imp_;
    };

    /// @param[out]  numGroups  The number of groups defined in the database.
    ///
    bool getNumTranslationGroups( int &numGroups );

    /// @param[in]   index  Index of the translation group to fetch
    /// @param[out]  group  TranslationGroup class describing the translation group
    ///
    bool getTranslationGroup( int index, TranslationGroup &group );

    /// @brief This class is used to retrieve information of a host translation defined
    /// in the path translation database.
    class WIRETAP_EXPORT_SPECIFIER HostTranslation
    {
    public:
        HostTranslation();
        ~HostTranslation();
        HostTranslation & operator=( const WireTapHostTranslation & i);

        /// @return the source host being translated
        ///
        const char *getSourceHost() const;

        /// @return the path being translated
        ///
        const char *getSourcePath() const;

        /// @return the host the translation is being performed on
        ///
        const char *getDestinationHost() const;

        /// @return the translated path
        ///
        const char *getDestinationPath() const;

    private:
        WireTapHostTranslation *imp_;
    };

    /// @param[out]  numHostTranslations  the number of host translations defined in the database.
    ///
    bool getNumHostTranslations( int &numHostTranslations );

    /// @param[in]   index            Index of the host translation to fetch
    /// @param[out]  hostTranslation  HostTranslation class describing the translation
    ///
    bool getHostTranslation( int index, HostTranslation &hostTranslation );

    /// @brief This class is used to retrieve information of a operating system 
    /// type translation defined in the path translation database.
    class WIRETAP_EXPORT_SPECIFIER OsTypeTranslation
    {
    public:
        OsTypeTranslation();
        ~OsTypeTranslation();
        OsTypeTranslation & operator=( const WireTapOsTypeTranslation & i);

        /// @return the source operating system type being translated
        ///
        const char *getSourceType() const;

        /// @return the path being translated
        ///
        const char *getSourcePath() const;

        /// @return the operating system the translation is being performed on
        ///
        const char *getDestinationType() const;
  
        /// @return the translated path
        ///
        const char *getDestinationPath() const;

    private:
        WireTapOsTypeTranslation *imp_;
    };

    /// @param[out]  numOsTypeTranslations the number of type translations defined in the database.
    ///
    bool getNumOsTypeTranslations( int &numOsTypeTranslations );

    /// @param[in]   idx                Index of the operating system translation to fetch
    /// @param[out]  osTypeTranslation  OsTypeTranslation class describing the translation
    ///
    bool getOsTypeTranslation( int idx, OsTypeTranslation &osTypeTranslation );

    /// @brief Register an event handler to be called back when an event is
    /// received for a given subject.
    ///
    /// @param[in]  subject       Subject to monitor event on.
    /// @param[in]  handler       Handler object that will be called upon
    ///                           events reception.
    /// @param[in]  streamFormat  An input parameter describing the format
    ///                           of the metadata parameter
    /// @param[in]  metaData      An optional parameter describing server-
    ///                           specific event parameters
    ///
    /// @return true if registration worked.
    ///
    /// @note The same event handler can be used for multiple subjects
    ///
    /// @note Servers older than 2013.3 do not support events.
    ///
    /// @warning This function should not be called more than once
    ///          for a given subject.
    ///
    bool registerEventHandler( const WireTapNodeId & subject,
                               WireTapNodeEvent::Handler * handler,
                               const char* streamFormat = 0, 
                               const char* metaData = 0 ); 

    /// @brief Unregister an event handler registered on a given subject.
    ///
    /// @param[in]  subject     Subject that was monitored.
    /// @param[in]  handler     Handler object associated to the subject.
    ///
    /// @return true if unregistration worked.
    ///
    /// @note Servers older than 2013.3 do not support events.
    ///
    /// @warning This function should only be called for subjects that have
    ///          been previously registered.
    ///
    bool unregisterEventHandler( const WireTapNodeId & subject,
                                 WireTapNodeEvent::Handler * handler );

private:
    WireTapServerHandle& operator=( const WireTapServerHandle& );

private:
    mutable WireTapServerImp *imp_;
};

/// @brief This class is a handle to a particular node on the Wiretap server 
/// to which the Wiretap client is currently connected.  
///
/// A node represents a directory or file in the file system that is exposed by
/// the Wiretap server.
///
/// WireTapNodeHandle uses a node ID and a server handle to uniquely refer to 
/// the node. It provides a set of functions to interface with the node via the 
/// Wiretap protocol. Node information is read and cached within the node 
/// object. The Wiretap Client API is not a distributed object model. 
/// The information obtained using a node handles is not updated dynamically. 
/// To update the information, the relevant get method(s) must be called.
///
/// Nodes come in two flavors: nodes and clip nodes. A clip node is simply a 
/// specialization of node.
///
/// The Wiretap Client API deliberately dissociates itself from the extensive 
/// and dynamic list of node types provided by most proprietary database 
/// implementations (for example, project, library, reel, and so on). 
/// Wiretap base functionality is confined to generic node manipulation, and to
/// basic clip and frame I/O services. 
///
/// The interpretation of extended types as well as metadata is the 
/// responsibility of the Wiretap client. For example, all nodes have an 
/// extended type defined by the implementation of the underlying database 
/// on the Wiretap server. A Wiretap client that is aware of the proprietary 
/// implementation of the Wiretap server(s) with which it will communicate, 
/// can interpret these extended types to provide specific functionality. 
/// The vendor, product name, and version of a Wiretap server's database 
/// implementation can be obtained from the instance of WireTapServerHandle that
/// represents the connection to the server. In this way, specific products can 
/// coordinate private interoperability efforts among themselves, or define 
/// public interoperability standards.
///
class WIRETAP_EXPORT_SPECIFIER WireTapNodeHandle
{
public:

    /// @brief Constructs a new WireTapNodeHandle object with no member 
    /// variables set. 
    ///
    WireTapNodeHandle();

    /// @brief Copy constructor. 
    ///
    /// Constructs a new WireTapNodeHandle object that duplicates the content 
    /// of the original. 
    ///
    /// @param  nodeHandle  An input parameter. The WiretapNodeHandle object 
    ///                     that is to be copied.
    /// 
    WireTapNodeHandle( const WireTapNodeHandle& nodeHandle);
    
    /// @brief Assignment operator
    ///
    /// Initializes a WireTapNodeHandle object that duplicates the content 
    /// of the original. 
    ///
    /// @param  nodeHandle  An input parameter. The WiretapNodeHandle object 
    ///                     that is to be copied.
    /// 
    WireTapNodeHandle& operator=( const WireTapNodeHandle& nodeHandle );

    /// @brief Constructs a new WireTapNodeHandle object using references to  
    /// instances of WireTapServerHandle and WireTapNodeId.
    ///
    /// Wiretap clients may need to use this constructor if the storage device
    /// on which the node is located has been switched from one Wiretap server
    /// to another. In this situation, a Wiretap client that has retained the 
    /// storage ID along with the node ID, can do the following: 
    /// <ol>
    ///   <li>Call WireTapServerList.resolveNodeId to obtain the
    ///       server ID for the new server
    ///   <li>Construct a new WireTapServerHandle from the server ID.  
    ///   <li>Construct a new WireTapNodeHandle using this constructor. 
    /// </ol>    
    ///
    /// @param server  An input parameter. A handle to the server on which 
    ///                the node is located. 
    /// 
    /// @param nodeId  An input parameter. A WireTapNodeId object that contains
    ///                the ID of the node for which a handle is to be
    ///                constructed. 
    /// 
    WireTapNodeHandle( const WireTapServerHandle &server,
                       const WireTapNodeId &nodeId );

    /// @brief Constructs a new WireTapNodeHandle object using a reference to  
    /// a WireTapServerHandle object and a string that contains a node ID.
    ///
    /// Wiretap clients may need to use this constructor if the storage device
    /// on which the node is located has been switched from one Wiretap server
    /// to another. In this situation, a Wiretap client that has retained the 
    /// storage ID and has a string containing a node ID, can do the following: 
    /// <ol>
    ///   <li>Call WireTapServerList.resolveNodeId to obtain the
    ///       server ID for the new server
    ///   <li>Construct a new WireTapNodeHandle using this constructor. 
    /// </ol>    
    ///
    /// @param server     An input parameter. A handle to the server on which 
    ///                   the node is located. 
    /// 
    /// @param nodeIdStr  An input parameter. A string that contains the ID
    ///                   of the node for which a handle is to be constructed. 
    /// 
    WireTapNodeHandle( const WireTapServerHandle &server,
                       const char *nodeIdStr );

    /// @brief Destructor.
    ///
    virtual ~WireTapNodeHandle();

    /// @brief Gets the persistent identifier of the current node.  
    ///
    /// @note Normally, a node ID is only valid in the context of the 
    /// WireTapServerHandle that was connected to the Wiretap server when the 
    /// node was created. After a Wiretap clients gets the node ID, the storage
    /// device (on which the node is actually stored) could be switched to a 
    /// different Wiretap server host. To be able to use node IDs at any time 
    /// (even if a storage device on which the node resides is switched to a 
    /// different host), Wiretap clients should retain the storage ID for all 
    /// nodes when the nodes are created. The storage ID can be obtained by 
    /// calling WireTapServerHandle.getStorageId. In case of difficulty 
    /// accessing a node using its node ID, Wiretap clients can call 
    /// WireTapServerList.resolve to obtain the coordinates of the 
    /// Wiretap server host that corresponds to the retained storage ID, and 
    /// then create a new server handle for the new Wiretap server host.
    ///
    /// @return  A WireTapNodeId object that contains the persistent identifier
    /// of the node. 
    ///
    WireTapNodeId &getNodeId() const;

    /// @brief Sets the persistent identifier of the current node.  
    ///
    /// @see WireTapNodeHandle.getNodeId.
    ///
    /// @param  nodeId  A WireTapNodeId object that contains the persistent 
    ///                 identifier of the node. 
    ///
    void setNodeId( const WireTapNodeId &nodeId );

    /// @brief Returns the parent node of the current node.
    ///
    /// @param parent  An output parameter. The method fills this parameter with
    ///                an instance of WireTapNodeHandle for the parent node.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getParentNode( WireTapNodeHandle &parent ) const;
    
    /// @brief Gets the number of child nodes found under the current node.  
    ///
    /// @param numChildren An output parameter. The method fills this parameter
    ///                    with the number of child nodes of the current node. 
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getNumChildren( unsigned &numChildren ) const;
    WIRETAP_DEPRECATED bool getNumChildren( int &numChildren ) const;

    /// @brief Gets a child node of the current node by specifying the child's 
    /// index.  
    ///
    /// The total number of child nodes for the current node is obtained by
    /// calling WireTapNodeHandle.getNumChildren.
    ///
    /// @param index  An input parameter. Node indexes start at 0. An index less
    ///               than zero or equal to the total number of children 
    ///               (obtained by calling WireTapNodeHandle.getNumChildren) 
    ///               will cause an error.
    ///
    /// @param child  An output parameter. The method fills this parameter with
    ///               an instance of WireTapNodeHandle for the node at the 
    ///               specified index.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getChild( int index, WireTapNodeHandle &child ) const;
 
    /// @brief Creates a node using an extended type defined by a 
    /// Wiretap server.  
    /// 
    /// This is the preferred method for instantiating any node 
    /// that is not a clip. The other method for creating non-clip nodes 
    /// (using an integer for the extended type) is deprecated.  
    ///
    /// To create a clip node, use 
    ///     WireTapNodeHandle.createClipNode( const char *displayName,
    ///                 const &WireTapClipFormat &f, const char *extendedType,
    ///                 &WireTapNodeHandle &node_return  ).
    ///
    /// @param  displayName   An input parameter. The name to use as a label on 
    ///                       the node in a user interface. Display names do not
    ///                       have to be unique.
    ///
    /// @param  extendedType  An input parameter. A string that is one of the 
    ///                       extended types defined by the Wiretap server on 
    ///                       which the node is to be created.
    ///
    /// @param  nodeHandle   An output parameter. The method fills this 
    ///                      parameter with an instance of WireTapNodeHandle 
    ///                      with the specified display name and extended type.
    ///                 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    /// 
    bool createNode( const char *displayName, const char *extendedType,
                     WireTapNodeHandle &nodeHandle );
    
    /// @brief Creates a node using an extended type defined by a 
    /// Wiretap server and initializes it with the supplied metadata  
    /// 
    /// This is the preferred method for instantiating any node 
    /// that is not a clip. The other method for creating non-clip nodes 
    /// (using an integer for the extended type) is deprecated.  
    ///
    ///
    /// @param  displayName   An input parameter. The name to use as a label on 
    ///                       the node in a user interface. Display names do not
    ///                       have to be unique.
    ///
    /// @param  extendedType  An input parameter. A string that is one of the 
    ///                       extended types defined by the Wiretap server on 
    ///                       which the node is to be created.
    ///
    /// @param  streamFormat  An input parameter. A string that describes the 
    ///                       the format of the stream. For example, to set
    ///                       XML format metadata for a project or user node on 
    ///                       an IFFFS Wiretap server, this parameter should be 
    ///                       set to "XML". @see setMetaData 
    ///
    /// @param  metaData      An input parameter. An ASCII string that defines 
    ///                       the metadata to associate with the node. 
    ///                       @see setMetadata
    ///
    /// @param  nodeHandle   An output parameter. The method fills this 
    ///                      parameter with an instance of WireTapNodeHandle 
    ///                      with the specified display name and extended type.
    ///                 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    /// 
    bool createNode( const char *displayName, const char *extendedType, 
                     const char* streamFormat, const char* metaData, 
                     WireTapNodeHandle &nodeHandle );

    /// @brief Creates a clip node using an extended type defined
    /// by a Wiretap server.  
    ///                 
    /// This is the preferred method for instanting a clip node.
    /// The other method for creating clip nodes (using an integer for the 
    /// extended type) is deprecated.  
    ///
    /// @note The sample program createClip.cpp supplied with the SDK shows how 
    /// to use this method to create a clip node.
    ///
    /// To create a node that is not a clip node, see 
    /// WireTapNodeHandle.createNode
    ///
    /// @param  displayName   An input parameter. The name to use as a label on 
    ///                       the node in a user interface. Display names do not
    ///                       have to be unique.
    ///
    /// @param  clipFormat    An input/output parameter. On input, a Wiretap 
    ///                       client should supply an instance of 
    ///                       WireTapClipFormat whose properties are set to 
    ///                       appropriate values. On output, the Wiretap server 
    ///                       will have set certain properties such as buffer 
    ///                       size.
    /// 
    /// @param  extendedType  An input parameter. A string that is one of the 
    ///                       extended types defined by the Wiretap server on 
    ///                       which the node is to be created.
    ///
    /// @param  nodeHandle   An output parameter. The method fills this 
    ///                      parameter with an instance of WireTapNodeHandle 
    ///                      with the specified display name, clip format, 
    ///                      and extended type.
    ///                 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///                      
    bool createClipNode( const char *displayName,
                         const WireTapClipFormat &clipFormat,
                         const char *extendedType,
                         WireTapNodeHandle &nodeHandle );

    /// @brief Gets if an ASCII metadata stream is associated with the
    /// current node.
    ///
    /// A Wiretap client can determine if a given metadata stream is available
    /// with this method. 
    ///
    /// @param  extendedType  An input parameter. A string that is one of the 
    ///                       extended types defined by the Wiretap server on 
    ///                       which the node is to be created.
    ///
    /// @param  canCreate     An output parameter. The method fills this
    ///                       parameter to true if the extended type is a valid
    ///                       type of child for the current node.
    ///
    /// @return  true if it was possible to get validity of the extended type
    /// for the current node, false if it was not possible to get the validity
    /// of the extended type for the current node. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string.
    ///
    bool canCreateNode( const char *extendedType,
                        bool &canCreate ) const;

    /// @brief Renames the node associated with the current node handle.
    /// 
    /// This method can be used to change the display name of a node.
    ///
    /// @note This may have the effect of changing the node's id.
    ///
    /// @param  displayName  An input parameter. The name to use as a label on 
    ///                      the node in a user interface.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    /// 
    bool renameNode( const char *displayName );
    
    /// @brief Destroys the node associated with the current node handle.  
    ///                 
    /// @return  true on success, false on failure. On success, the node is no 
    /// longer in the file system of the Wiretap server, the current node handle
    /// is no longer valid, and function calls on the node handle generate 
    /// errors. On failure, WireTapNodeHandle::lastError() can be called to get 
    /// the error string.
    ///
    bool destroyNode();

    /// @brief Creates a duplicate of the supplied node and adds it to this node
    ///
    /// @param  dupeNode      The node to be duplicated
    ///
    /// @param  displayName   The display name to apply to the duplicated node
    ///
    /// @param  newNode       A reference to the new, duplicated node
    ///
    /// @return  true on success, false on failure. On failure,
    /// WireTapNodeHandle::lastError() can be called to get the error string.
    ///
    bool duplicateNode( const WireTapNodeHandle& dupeNode,
                        const char *displayName,
                        WireTapNodeHandle &newNode );

    /// @brief Determines whether the current node is a clip node.  
    ///        
    /// @param  isClipNode  An output parameter. The method sets this parameter
    ///                     to true if the node is a clip node, or sets it to 
    ///                     false if the node is not a clip node.
    ///         
    /// @return  true if the method is able to determine whether the node is a
    /// clip node, false if the method cannot determine whether the node is a
    /// clip node. On failure, WireTapNodeHandle::lastError() can be called to 
    /// get the error string.
    ///
    bool getIsClipNode( bool &isClipNode ) const;

    /// @brief Gets the type of the current node expressed as an integer.  
    ///        
    /// @param  type  An output parameter. The method sets this parameter
    ///               to the integer value for the extended node type. 
    ///               Each Wiretap server defines its own extended node types. 
    ///               The default value (for a node with no extended type) is 0.
    ///         
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    /// @deprecated Since Wiretap 2007.1.  The preferred method is 
    /// WireTapNodeHandle.getNodeTypeStr.
    ///
    WIRETAP_DEPRECATED bool getNodeType( int &type ) const;


    /// @brief Gets a string that describes the type of the current node. 
    ///        
    /// @param  type  An output parameter. The method fills this parameter
    ///               with a WireTapStr object that gives access to the node 
    ///               type as a string. This could be the value entered when the
    ///               node was created or on of the extended node types defined 
    ///               by the Wiretap server. Each Wiretap server defines its own 
    ///               extended node types.         
    ///         
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getNodeTypeStr( WireTapStr &type ) const;

    /// @brief Gets the display name of the current node.  
    ///        
    /// @param  name  An output parameter. The method fills this parameter
    ///               with a WireTapStr object containing the display name for 
    ///               the node. Display names need not be unique. 
    ///         
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getDisplayName( WireTapStr &name ) const;

    /// @brief Gets the clip format associated with the current clip node.  
    ///        
    /// This method fails if the current node is not a clip node.
    ///
    /// @param  format  An output parameter. The method fills this parameter
    ///                 with a WireTapClipFormat object of the type set when the 
    ///                 node was created. 
    ///         
    /// @return  true if the node is a clip node and the format was found, 
    /// false on failure. On failure, WireTapNodeHandle::lastError() can be 
    /// called to get the error string. 
    ///
    bool getClipFormat( WireTapClipFormat &format ) const;

    /// @brief Gets ASCII metadata in a specified format for the current node.
    ///
    /// Wiretap clients can use this method directly if the desired type of 
    /// metadata is known (for example, "XML"). If the desired type of metadata 
    /// stream is not known or if there are multiple metadata streams, 
    /// the available metadata stream(s) for the current node can be 
    /// iterated through and obtained with a combination of methods:
    /// <ul>
    ///   <li>WireTapNodeHandle.getNumAvailableMetaDataStreams
    ///   <li>WireTapNodeHandle.getAvailableMetaDataStream
    ///   <li>this method (getMetaData)
    /// </ul> 
    ///
    /// Not all nodes have ASCII metadata associated with them. See the 
    /// Wiretap Developer's Guide for more information about the types of nodes 
    /// that support metadata and which metadata formats they support.  
    ///        
    /// @param  streamFormat  An input parameter. A string that describes the 
    ///                       format of the stream. For example, to get
    ///                       XML format metadata for a project or user node on
    ///                       an IFFFS Wiretap server, this parameter should be
    ///                       set to "XML". 
    ///         
    /// @param  filter        An input parameter. A string that is the name of 
    ///                       an element in the metadata stream. 
    ///                       For some node types, it is possible to select 
    ///                       individual elements in the metadata stream. 
    ///                       See the Wiretap Developer's Guide for more 
    ///                       information. 
    ///                       To get the entire metadata stream, set this  
    ///                       parameter to an empty string (""). 
    ///
    /// @param  depth         An input parameter. An integer that specifies how
    ///                       many levels to recurse within the metadata 
    ///                       structure. Not all nodes or metadata streams 
    ///                       support this parameter. See the Wiretap
    ///                       Developer's Guide for more information.
    ///         
    /// @param  metaData      An output parameter. The method fills this 
    ///                       parameter with a WireTapStr object that contains 
    ///                       the metadata for the node as an ASCII string.
    ///
    /// @return  true if it was possible to get the metadata specified, 
    /// false if it was not possible to get the metadata. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getMetaData( const char *streamFormat, const char *filter,
                      const int depth, WireTapStr &metaData ) const;

    /// @brief Sets ASCII metadata for the current node in a specified format.
    ///
    /// Not all nodes can have ASCII metadata associated with them. See the 
    /// Wiretap Developer's Guide for more information about the types of nodes 
    /// that support metadata and which metadata formats they support.  
    ///        
    /// @param  streamFormat  An input parameter. A string that describes the 
    ///                       the format of the stream. For example, to set
    ///                       XML format metadata for a project or user node on 
    ///                       an IFFFS Wiretap server, this parameter should be 
    ///                       set to "XML".  
    ///         
    /// @param  metaData      An input parameter. A WireTapStr object 
    ///                       containing an ASCII string that defines the 
    ///                       metadata to associate with the node.
    ///
    /// @return  true if it was possible to set the metadata as specified, 
    /// false if it was not possible to set the metadata. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool setMetaData( const char *streamFormat, const char *metaData ) const;

    /// @brief Gets the number of ASCII metadata
    /// streams that are associated with the current node.
    ///
    /// A Wiretap client can determine the number of metadata streams 
    /// that are available with this method, and then it can iterate and obtain
    /// the stream types (using the WireTapNodeHandle.getAvailableMetaDataStream
    /// method) and the actual metadata (using the WireTapNodeHandle.getMetaData
    /// method). 
    ///
    /// @param  numStreams  An output parameter. The method fills this parameter
    ///                     with the number of metadata streams that are 
    ///                     available for the current node.
    ///
    /// @return  true if it was possible to get the number of metadata streams, 
    /// false if it was not possible to get the number of streams. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getNumAvailableMetaDataStreams( int &numStreams ) const;

    /// @brief Gets the format of an ASCII metadata stream that is
    /// associated with the current node.
    ///
    /// After determining how many metadata streams are available 
    /// using the WireTapNodeHandle.getNumAvailableMetaDataStreams 
    /// method, a Wiretap client can iterate on the stream indexes to get the 
    /// formats of the metadata streams (using this method) or to get 
    /// the actual metadata (using WireTapNodeHandle.getMetaData).
    ///
    /// If the format of the metadata stream is already known, the actual stream
    /// can be obtained directly with the WireTapNodeHandle.getMetaData method. 
    ///        
    /// @param  index    An input parameter. The index of the metadata stream
    ///                  to be queried.
    ///
    /// @param  format   An output parameter. The method fills this parameter
    ///                  with a WireTapStr object that is the format of the
    ///                  metadata (for example, XML).
    ///
    /// @return  true if it was possible to get the metadata format, 
    /// false if it was not possible to get the format. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///
    bool getAvailableMetaDataStream( int index, WireTapStr &format ) const;

    /// @brief Gets if an ASCII metadata stream is associated with the
    /// current node.
    ///
    /// A Wiretap client can determine if a given metadata stream is available
    /// with this method. 
    ///
    /// @param  streamFormat  An input parameter. A string that describes the 
    ///                       the format of the stream. For example, to set
    ///                       XML format metadata for a project or user node on 
    ///                       an IFFFS Wiretap server, this parameter should be 
    ///                       set to "XML".
    ///
    /// @param  isAvailable An output parameter. The method fills this parameter
    ///                     to true if the stream is available for the current
    ///                     node.
    ///
    /// @return  true if it was possible to get availability of the metadata
    /// stream, false if it was not possible to get the availability of the
    /// stream. On failure,  WireTapNodeHandle::lastError() can be called to
    /// get the error string.
    ///
    bool getIsMetaDataStreamAvailable( const char *streamFormat,
                                       bool &isAvailable ) const;

    /// @brief Gets a string that is the identifier for a raw data
    /// stream associated with the node.
    ///
    /// @note Wiretap clients would typically call this method to obtain a
    /// stream ID for the node before calling WireTapServerHandle.readStream.
    ///
    /// @param  streamId  An output parameter. A string that is the identifier 
    /// for the raw data stream associated with the node. If the node does not 
    /// have a raw data stream associated with it, or if the Wiretap server does 
    /// not support raw data streams, an empty string ("") is returned. 
    ///
    /// @return  true if a WireTapStr object was updated with a valid value 
    /// (empty or not empty, see the streamId parameter above) is returned;
    /// false if it was not possible to update streamId. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    /// 
    bool getStreamId( WireTapStr &streamId ) const;

    enum ReadAheadHint
    {
      NO_READ_AHEAD     = 0,
      READING_FORWARD   = 1,
      READING_BACKWARD  = -1
    };

    /// @brief Reads the contents of a frame at a specified index 
    /// in the current clip node and writes it to a buffer.    
    ///
    /// This method fails if the current node is not a clip node.
    ///
    /// Wiretap clients can call WireTapNodeHandle.getNumFrames to determine the
    /// number of frames to be read from the node. Then they would iterate 
    /// through the frames and call this method to read the frames from the 
    /// buffer. The sample program readFrames.cpp (which is included in the 
    /// SDK) shows how to use this method.
    ///
    /// @note Use WireTapServerHandle.readFrame if only the frame ID is known
    /// but it is not associated with a clip node.
    ///
    /// @param  frameIndex  An input parameter. The index of the frame to be 
    ///                     read. Frame indexes start at 0. An index less than 
    ///                     zero or equal to the number of available frames 
    ///                     (obtained by calling WireTapNodeHandle.getNumFrames)
    ///                     will generate an error. 
    ///
    /// @param  params      An input parameter. Contains black data on the
    ///                     parameter to use to decode and read the frame.
    ///                     Might be 0 if none needed.
    ///
    /// @param  paramsSize  An input parameter. The number of bytes in the
    ///                     the buffer to be used as the read parameters.
    ///
    /// @param  buffer      An input parameter. The buffer to which the frame 
    ///                     will be written. This value cannot be NULL. 
    ///
    /// @param  bufferSize  An input parameter. The number of bytes in the  
    ///                     the buffer to be used for the frame. 
    ///                     The buffer size must be equal to or greater than 
    ///                     the value obtained from the WireTapClipFormat 
    ///                     object for the node. The sample program 
    ///                     readFrames.cpp (included in the SDK) shows how to
    ///                     find the minimum buffer size.
    /// 
    /// @param  readAheadHint By default, the next frames in the clip sequence
    ///                       will be sent to the server as a hint to prepare
    ///                       (i.e. read ahead) and decode the next frame. 
    ///                       Setting this parameter to NO_READ_AHEAD will
    ///                       disable this functionality.  This option would
    ///                       normally be used when generating a thumbnail.
    ///                       Setting this parameter to READING_BACKWARD will
    ///                       resverse the direction of the 'next frame'.
    /// 
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    /// 
    bool readFrame( int frameIndex,
                    const void * params, size_t paramsSize,
                    void * buffer, size_t bufferSize,
                    ReadAheadHint readAheadHint = READING_FORWARD );

    /// @deprecated Since Wiretap 2010.1.1.  The preferred method is 
    /// WireTapNodeHandle.readFrame() with params and readAheadHint.
    ///
    WIRETAP_DEPRECATED bool readFrame( int frameIndex,
                                       void *buffer, size_t bufferSize,
                                       bool noReadAhead = false );

    /// @brief Reads the contents of a buffer and writes it to a frame at a 
    /// specified index in the current clip node.
    ///
    /// This method fails if the current node is not a clip node.
    ///
    /// Wiretap clients can call WireTapNodeHandle.getNumFrames to obtain the 
    /// number of frames in the clip node. Then they would this 
    /// method to write the frames to the buffer. The sample program 
    /// createClip.cpp (which is included in the SDK) shows how to use this 
    /// method.
    ///
    /// @note Use WireTapServerHandle.writeFrame can be used if only the frame 
    /// ID is known but it is not associated with a node. 
    ///
    /// @param  frameIndex  An input parameter. The index of the frame to be 
    ///                     written.
    ///
    /// @param  buffer      An output parameter. The buffer from which the frame
    ///                     will be read. This value cannot be NULL.
    ///
    /// @param  bufferSize  An input parameter. The number of bytes in the  
    ///                     the buffer to be used for the frame. 
    ///                     The buffer size must be equal to or greater than 
    ///                     the value obtained from the WireTapClipFormat 
    ///                     object for the node. The sample program 
    ///                     createClip.cpp (included in the SDK) shows how to
    ///                     find the minimum buffer size.
    ///
    /// @return  true on success, false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///  
    bool writeFrame( int frameIndex, const void *buffer, size_t bufferSize );

    /// @brief  Gets the identifier of a frame associated with the 
    /// current clip node.
    ///
    /// This method fails if the current node is not a clip node.
    ///        
    /// The total number of frames in the clip can be obtained with the method 
    /// WireTapNodeHandle.getNumFrames. 
    ///
    /// The frame ID (obtained with this method) and the frame file path
    /// (obtained with WireTapNodeHandle.getFrameIdPath) allow Wiretap clients 
    /// to access the raw frame directly rather than through the Wiretap server 
    /// TCP connection. Depending on the bandwidth of the direct connection to 
    /// the storage device, a significant gain in performance can be achieved.
    ///
    /// @param  frameIndex  An input parameter. The index of the frame whose ID
    ///                     is being sought. The total number of frames in the
    ///                     clip is obtained with the method 
    ///                     WireTapNodeHandle.getNumFrames.
    ///
    /// @param  frameId     An output parameter. The method fills this method
    ///                     with a WireTapStr object that contains the 
    ///                     frame identifier.
    ///
    /// @return true if the node is a clip node, and it was possible to get the 
    /// frame ID; false if the node is not a clip node, or it was not
    /// possible to get the frame ID. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string. 
    ///  
    bool getFrameId( int frameIndex, WireTapStr &frameId ) const;

    /// @brief  Gets the file path of a frame in the current clip node.
    ///
    /// This method fails if the current node is not a clip node.
    ///
    /// The total number of frames in the clip can be obtained with the method 
    /// WireTapNodeHandle.getNumFrames. 
    ///
    /// The frame ID (obtained with WireTapNodeHandle.getFrameId) and the file 
    /// path (obtained with this method) allow clients to access the raw frame
    /// directly rather than through the Wiretap server TCP connection. 
    /// Depending on the bandwidth of the direct connection to the storage 
    /// device, a significant gain in performance can be achieved.
    ///
    /// @param  frameIndex  An input parameter. The index of the frame whose 
    ///                     frame path is sought. 
    ///
    /// @param  path        An output parameter. The method fills this method
    ///                     with a WireTapStr object that contains the 
    ///                     file path of the frame. If the file path is not 
    ///                     available, the path will be set to the empty string.
    ///
    /// @return  true if the path variable was set to set to a valid path or to
    /// an empty string; false on failure. On failure, 
    /// WireTapNodeHandle::lastError() can be called to get the error string.
    ///  
    bool getFrameIdPath( int frameIndex, WireTapStr &path ) const;

    /// @brief Preallocates the number of frames in the current clip node.
    ///
    /// This method fails if the current node is not a clip node.
    ///
    /// @note The sample program createClip.cpp shows how to use this method.
    ///
    /// @param numberOfFrames An input parameter. The number of frames to
    ///                       allocate for the clip.   
    ///
    /// @return true if the node is a clip node, and it was possible to set the 
    /// number of frames; false if the node is not a clip node, or it was not
    /// possible to set the number of frames. On failure,
    /// WireTapNodeHandle::lastError() can be called to get the actual error 
    /// string.
    ///
    bool setNumFrames( unsigned numberOfFrames );
 
    /// @brief Gets the number of frames of raw data associated with the current
    /// node, if it is a clip node.  
    ///
    /// @param numberOfFrames An output parameter. The number of frames that
    ///                       are in the clip.   
    ///
    /// @return true if the node is a clip node, and it was possible to get the 
    /// number of frames; false if the node is not a clip node, or it was not
    /// possible to get the number of frames. On failure,
    /// WireTapNodeHandle::lastError() can be called to get the actual error 
    /// string.
    ///
    bool getNumFrames( unsigned &numberOfFrames ) const;
    WIRETAP_DEPRECATED bool getNumFrames( int &numberOfFrames ) const;

    /// @brief Links to specified file paths in order to soft-import frames to 
    /// the current clip node.  
    ///
    /// This method fails if the current node is not a clip node.
    /// 
    /// @note The sample program createSoftClip.cpp shows how to use this
    ///       method.
    ///
    /// The clip node must be empty when this method is called.  The frame
    /// formats of the individual frame files must match the clip format of the
    /// clip node or an error will be generated.
    /// The frame paths are automatically translated to use the native mount
    /// point of the server, using the Wiretap path translation services.
    ///
    /// @param  paths          An input parameter. A WireTapStr object that 
    ///                        points to an array of paths for the frame files 
    ///                        to be linked to. 
    ///
    /// @param numberOfPaths   An input parameter. The number of frame paths 
    ///                        that are in the paths parameter.
    ///
    /// @return true if the node is a clip node, and it was possible to link to
    /// frames; false if the node is not a clip node, or it was not
    /// possible to link to the frames. On failure,
    /// WireTapNodeHandle::lastError() can be called to get the actual error 
    /// string.
    ///
    bool linkToFrames( const WireTapStr *paths, int numberOfPaths );

    /// @brief Gets the error string for the last operation that failed.  
    ///
    /// Wiretap clients can obtain this error message for any member function
    /// that returns a boolean. The strings can be localized if required. 
    /// See the Wiretap Developer's Guide. 
    ///
    /// @return The error string for the last operation (called on the current 
    /// node handle) that failed.
    const char *lastError() const;

    /// @brief Gets a reference to the WireTapServerHandle for the current node.
    ///
    /// A node handle and a node ID are only valid in the context of the server
    /// under which they were created.
    /// A node's server is normally set when the node handle is created 
    /// (by methods such as WireTapNodeHandle.createNode, 
    /// WireTapNodeHandle.createClipNode, WireTapServerHandle.getRootNode).
    ///
    /// @return A reference to the WireTapServerHandle for the current node.
    ///
    const WireTapServerHandle &getServer() const;

    /// @brief Sets the WireTapServerHandle for the current node.  
    ///
    /// A node handle and a node ID are only valid in the context of the server
    /// under which they were created.
    /// A node's server is normally set when the node handle is created 
    /// (by any number of methods such as WireTapNodeHandle.createNode, 
    /// WireTapNodeHandle.createClipNode, WireTapServerHandle.getRootNode,
    /// WireTapServerList.getNode). Wiretap clients should not need to call this
    /// method in normal circumstances.
    ///
    /// @param  server  An input parameter. A reference to the 
    ///                 WireTapServerHandle for the current node.
    ///
    void setServer( const WireTapServerHandle &server );


    /// @note activateReadAhead is not required under normal circumstances.
    ///       It should not be used in general purpose code.
    ///
    bool activateReadAhead( );
    bool getNbReadAheads( int &nbReadAheads ) const;

public:
    /// @brief Creates a node using an integer type defined by the IFFFS 
    /// Wiretap server.  
    ///
    /// @deprecated Since Wiretap 2007.1. The preferred method is 
    /// WireTapNodeHandle.createNode( const char *displayName, 
    /// const char *extendedType, %WireTapNodeHandle &nodeHandle )
    ///                      
    WIRETAP_DEPRECATED bool createNode( const char *displayName,
                                        int extendedType,
                                        WireTapNodeHandle &nodeHandle );

    /// @brief Creates a clip node using an integer type defined by the IFFFS 
    /// Wiretap server.  
    ///
    /// @deprecated Since Wiretap 2007.1. The preferred method is 
    /// WireTapNodeHandle.createClipNode( const char *displayName,
    ///   const %WireTapClipFormat &format, const char *extendedType,
    ///   %WireTapNodeHandle &nodeHandle  )
    ///                      
    WIRETAP_DEPRECATED bool createClipNode( const char *displayName,
                                            const WireTapClipFormat &clipFormat,
                                            int extendedType,
                                            WireTapNodeHandle &nodeHandle );

private:
    WireTapNodeImp &imp() const;

private:
    WireTapNodeImp *imp_;
};

/// @brief Gets a handle for the node with a specified display name that is the 
/// child of a specified parent node.
///
/// This function gets a node whose display name matches the name specified as 
/// an input parameter. The node is the first child node with that display name
/// that is found under the parent node (also specified as an input parameter).
/// Since Wiretap does not prevent several nodes from having the same display 
/// name, this method does not necessarily always find the same node. 
/// @note The uniqueness of a node on a given server is determined by its node 
/// ID not its display name. The node ID should be used as the persistent 
/// reference to a node (rather than a display name).
///
/// @param parentNode  An input parameter. The node whose child is being sought.
///
/// @param displayName An input parameter. The displayName of the node sought. 
///                    This parameter is case-sensitive. This method verifies 
///                       the display name of each child node until it finds one
///                    that matches.
///
/// @param child       An output parameter. The method places an instance of 
///                    WireTapNodeHandle representing the child node in this 
///                    parameter. 
///
/// @return  true if a child node with the specified name was found for the 
/// parentNode; false if the parent node does not have any children 
/// with the specified display name.
///
WIRETAP_EXPORT_SPECIFIER
bool WireTapFindChild( const WireTapNodeHandle &parentNode,
                       const char *displayName,
                       WireTapNodeHandle &child );

/// @brief Gets a node handle that corresponds to the specified display name on
/// the specified Wiretap server. 
///
/// Gets a node handle object that represents the node at the specified path on
/// the specified Wiretap server. The specified path terminates with the display
/// name of the node. The search is case-sensitive and does not handle 
/// wildcards.
/// 
/// @param server      An input parameter. A Wiretap server. The method will
///                    attempt to resolve the path (specified in the displayPath
///                    parameter) on this server.
/// @param displayPath An input parameter. A fully or partially qualified 
///                    path to a node on the server. The method will attempt to 
///                    resolve this path. displayPath must be formed as follows:
///                    <ul>
///                      <li>displayPath must end with the node display name. 
///                      <li>Valid display names for nodes do not contain 
///                          forward or backward slash characters.
///                      <li>The path separator can be either a forward or a 
///                          backward slash.
///                      <li>P ath separators can be doubled. For example: "//"
///                      <li>The path can have leading and trailing slashes.
///                      <li>Mixing of forward and backward slashes is legal.
///                      <li>For relative paths: 
///                      <ul> 
///                          <li>"/./" is supported. 
///                          <li>"/../" not supported.
///                        </ul>
///                    </ul>
///
/// @param child       An output parameter. After a successful search, this 
///                    parameter contains an instance of WireTapNodeHandle that 
///                    points to the child node.
///
/// @return  true if a node with the specified path was found on the specified 
/// server; false if the server does not have a node with the 
/// specified path or if the specified path is not well-formed.
///
WIRETAP_EXPORT_SPECIFIER bool WireTapResolveDisplayPath(
                                           const WireTapServerHandle &server,
                                           const char *displayPath,
                                           WireTapNodeHandle &child );

/// @class WireTapServerList
/// @brief This class prepares a list of the Wiretap servers that can be 
/// accessed by the Wiretap client.
///
/// This class sends out a multicast request on the local subnet. Wiretap 
/// clients can use this class to populate a browser with a list of the Wiretap 
/// servers available on a host or network.
/// <p>The list can be filtered to include the servers associated with a 
/// particular database which, in effect, means a particular type of Wiretap 
/// server (IFFFS, Backburner, and so on). The class can also prepare 
/// a list of Wiretap servers that are accessible via a particular gateway. 
/// @note This class is not thread safe.
///
class WIRETAP_EXPORT_SPECIFIER WireTapServerList
{
public:
    /// @brief The constructor prepares a list of the servers on the subnet to 
    /// which the Wiretap client is connected. 
    ///
    /// The constructor sends out a multicast request that reaches a single 
    /// server from which all other servers are, in turn, located. By default, 
    /// all servers that respond to the multicast message are added to the list.
    /// Two optional parameters can be set to customize the list.
    ///
    /// @param database      An input parameter. This optional parameter can be 
    ///                      set to the database identifier for a particular 
    ///                      Wiretap server. Since database IDs are 
    ///                      product-specific and unique across vendors, setting
    ///                      this parameter makes it possible to filter for a 
    ///                      particular type of Wiretap server. Examples of
    ///                      these database IDs include "IFFFS", "Backburner". 
    ///                      If this parameter is not set, all servers on the 
    ///                      subnet will be added to the list.
    ///
    /// @param gatewayIpAddr An input parameter. This optional parameter can be 
    ///                      set to the IP address of the gateway computer to 
    ///                      which the multicast message should be sent. If this
    ///                      parameter is not set, the servers on the subnet 
    ///                      where the Wiretap client is running will be 
    ///                      listed.
    ///
    /// @param gatewayPort   An input parameter. This optional parameter can be 
    ///                      set to the port of the gateway server to which the 
    ///                      multicast message should be sent. If this parameter
    ///                      is not set, the servers on the subnet where the 
    ///                      Wiretap client is running will be listed.
    ///
    WireTapServerList( const char *database=0,
                       const char *gatewayIpAddr=0,
                       int gatewayPort=0 );

    /// @brief Destructor.
    ///
    ~WireTapServerList();

    /// @brief Resolves server ID to its corresponding server information.
    ///
    /// This method retrieves a WireTapServerInfo object based on a
    /// WireTapServerId object containing partial information that defines
    /// a persistent identifier for the Wiretap server (an instance of a 
    /// particular Wiretap server running on a particular host machine).
    /// Wiretap clients can use this WireTapServerId object to instantiate
    /// a WireTapServerHandle which gives access to the data on the 
    /// storage device.
    ///
    /// @see WireTapServerID, WireTapServerInfo, WireTapServerHandle.
    ///
    /// @param serverId   An input parameter. The WireTapServerId object
    ///                   for the Wiretap server ID to resolve.
    /// @param serverInfo An output parameter. The WireTapServerInfo object
    ///                   for the targeted Wiretap server ID.
    ///
    /// @return true if server ID was resolved successfully; false if the
    /// provided information was unsufficient or if the Wiretap server described
    /// by the server id is no longer active and broadcasting its presence.
    /// On failure, WireTapServerList::lastError() can be called to get the 
    /// error string.
    ///
    bool resolve( const WireTapServerId & serverId,
                  WireTapServerInfo & serverInfo );

    /// @brief Resolves the storage ID of a storage device to determine 
    /// which Wiretap server it is currently connected to. 
    ///
    /// This method obtains a WireTapServerId object containing the information 
    /// that defines a persistent identifier for the Wiretap server (an instance
    /// of a particular Wiretap server running on a particular host machine). 
    /// Wiretap clients can use this WireTapServerId object to instantiate a 
    /// WireTapServerHandle which gives access to the data on the 
    /// storage device.
    ///
    /// @see WireTapServerID, WireTapServerInfo, WireTapServerHandle.
    ///
    /// @param storageId  An input parameter. The storage ID of a 
    ///                   storage device.   
    ///
    /// @param serverId   An output parameter. The WireTapServerId object for
    ///                   the Wiretap server to which the storage device is 
    ///                   connected. 
    ///
    /// @return true if storageId was resolved successfully; false if the 
    /// the storage device is not connected to an active Wiretap server. 
    /// On failure, WireTapServerList::lastError() can be called to get the 
    /// error string.
    ///
    /// @deprecated Since Wiretap 2013.1. The preferred function is
    /// resolve( const WireTapServerId & serverId ) 
    ///
    WIRETAP_DEPRECATED bool resolveStorageId( const char *storageId,
                                              WireTapServerId &serverId );

    /// @brief Gets the number of Wiretap servers available to the
    /// Wiretap client.  
    ///
    /// This method determines the number of servers available on the network 
    /// that was specified when the WireTapServerList was created. If the
    /// server list was not filtered in any way, the default network is the one 
    /// accessible through the Wiretap client's local host. 
    /// Wiretap clients can iterate through the available servers and 
    /// use the getNode method to access each server.
    ///
    /// @param numberOfNodes  An output parameter. The number of Wiretap servers
    ///                       that are available on the network.   
    ///
    /// @return true if it was possible to get the number of servers available; 
    /// false if it was not possible to get the number. On failure,
    /// WireTapServerList::lastError() can be called to get the actual error 
    /// string.
    ///
    bool getNumNodes( int &numberOfNodes );  

    /// @brief Gets the WireTapServerInfo object at the specified index.  
    /// 
    /// Wiretap clients can use this method when iterating through the Wiretap
    /// servers in the list. The getNumNodes method obtains the number of server
    /// available on the network.
    ///
    /// The WireTapServerInfo object that is obtained by this method 
    /// contains all the information that a Wiretap client might want to display
    /// about a Wiretap server. It also gives access to the WireTapServerId 
    /// for the server which can used to obtain a WireTapServerHandle 
    /// for the server. To access data on a particular server, Wiretap clients 
    /// must obtain a WireTapServerHandle for that server.
    ///
    /// @see WireTapServerInfo, WireTapServerId, WireTapServerHandle
    ///
    /// @param index  An input parameter. The index of the Wiretap server in the
    ///               list of available servers.
    ///
    /// @param node   An output parameter. The WireTapServerId object that can
    ///               be used to obtain a WireTapServerHandle for the server.
    ///
    /// @return  true if it was possible to get a WireTapServerId object for
    /// specified index, false if it was not possible to get a WireTapServerId 
    /// object. On failure, WireTapServerList::lastError() can be called to get 
    /// the error string.
    ///
    bool getNode( int index, WireTapServerInfo & node );

    /// @brief Actively trigger a refresh of available servers on the network.
    ///
    /// Servers activity on the network is monitored to avoid the need to call
    /// this method. However, some network or OS condition might prevent
    /// network multicast message to be sent or received causing the list to
    /// be incorrect.
    ///
    /// Calling this method will brodcast a request to find new server on the
    /// network.
    ///
    /// @return true if operation was successful.
    /// On failure, WireTapServerList::lastError() can be called to get the 
    /// error string.
    ///
    bool refresh();

    /// @brief Gets the error string for the last operation that failed.  
    ///
    /// Wiretap clients can obtain this error message for any member function
    /// that returns a boolean. The strings can be localized if required. 
    /// See the Wiretap Developer's Guide. 
    ///
    /// @return The error string for the last operation (called on the current 
    /// server list) that failed.
    ///
    const char *lastError() const;

    /// @brief Gets the IP address of the gateway server from which the
    ///        multicast reply have been receive from.
    ///
    const char * getGatewayIpAddr() const;

    /// @brief Gets the port of the gateway server from which the
    ///        multicast reply have been receive from.
    ///
    int getGatewayPort() const;

    /// @brief This type definition is available for backward compatibility 
    /// with Wiretap 2007 and earlier versions.
    ///
    typedef WireTapServerInfo ServerInfo;

    WireTapServerList( const WireTapServerList & );
    WireTapServerList & operator = ( const WireTapServerList & );

private:
    #ifndef WIRETAP_STRIP_FROM_DOXYGEN
    class Impl;
    Impl * impl_;
    #endif
};

#endif
