#!/usr/bin/env python

##*****************************************************************************
##
## File: createTimeline.py
##
## Copyright (c) 2016 Autodesk, Inc.
## All rights reserved.
##
## Use of this software is subject to the terms of the Autodesk license
## agreement provided at the time of installation or download, or which
## otherwise accompanies this software in either electronic or hard copy form.
##*****************************************************************************

"""
Wiretap SDK sample python program to create an IFFFS timeline
from 3 sources clip nodes.
"""

from __future__ import print_function
from __future__ import absolute_import

import sys
import getopt

from adsk.libwiretapPythonClientAPI import WireTapClient
from adsk.libwiretapPythonClientAPI import WireTapServerHandle
from adsk.libwiretapPythonClientAPI import WireTapNodeHandle
from adsk.libwiretapPythonClientAPI import WireTapClipFormat
from adsk.libwiretapPythonClientAPI import WireTapStr

class WireTapException(Exception):
    pass

def create_clip_node(parent_node_handle,
                     display_name,
                     tape_name,
                     num_frames,
                     clip_node_handle):
    """
    Creation of a source clip node.
    """

    # Define an IFFS XML stream to provide the tape name to the
    # source clip that will be created.
    # Note that pre-2008 verions of the IFFFS Wiretap server accept
    # "IFFFS_XML" as the metadata stream name and root tag.
    #
    xml_stream = "<XML Version=\"1.0\">"
    xml_stream += "<ClipData>"
    xml_stream += "<TapeName>"
    xml_stream += tape_name
    xml_stream += "</TapeName>"
    xml_stream += "</ClipData>"
    xml_stream += "</XML>"

    print(xml_stream)

    # Define a clip format.  The parameters allow the server to
    # determine the frame buffer size for an NTSC raw RGB frame,
    # which can be queried when the clip is created.
    # The XML metadata tag allows you to specify the source tape
    # name, among other things.
    # Other information, such as the source timecode and the drop
    # mode, could be also provided to the source clip using the same
    # XML stream and XML metadata tag.
    #
    clip_format = WireTapClipFormat(
        720, 486,       # width, height
        3 * 8,          # bits per pixel
        3,              # # channels
        29.97,          # frame rate
        0.9,            # pixel ratio
        WireTapClipFormat.ScanFormat.SCAN_FORMAT_FIELD_1_ODD,
        WireTapClipFormat.FORMAT_RGB(),
        "XML",
        xml_stream
    )

    # Create the new clip node.  Each server will provide a list of
    # extended types (specifyable as a string or an enum).
    #
    if not parent_node_handle.createClipNode(
            display_name,       # display name
            clip_format,        # clip format
            "CLIP",             # extended (server-specific) type
            clip_node_handle):  # created node returned here
        raise WireTapException("Unable to create clip node: %s." % \
                               parent_node_handle.lastError())

    # Set the number of frames.
    #
    if not clip_node_handle.setNumFrames(num_frames):
        raise WireTapException("Unable to set the number of frames: %s." % \
                               clip_node_handle.lastError())

    print("Clip %s successfully created." % display_name)


def main(argv):

    # check the command line options for correctness
    try:
        opts, _ = getopt.getopt(argv, "h:n:")
    except getopt.GetoptError:
        print("Usage:\n" \
              "-h <hostname>\n" \
              "-n <node Id>")
        sys.exit(2)

    # default host name and node ID
    host_name = "localhost"
    node_id = ""

    # parse command line option to set specified host name
    # and node ID
    for opt, arg in opts:
        if opt == '-h':
            host_name = arg
        elif opt == '-n':
            node_id = arg

    # Initialize the Wiretap Client API.
    #
    wiretap_client = WireTapClient()
    if not wiretap_client.init():
        raise WireTapException("Unable to initialize Wiretap client API.")

    create_timeline(
        host_name=host_name,
        node_id=node_id
    )

def create_timeline(host_name, node_id):

    # Instantiate a server handle
    #
    server_handle = WireTapServerHandle(host_name)

    # Instantiate the clip node handles for the given server and the node ID
    #
    # To assemble a timeline from sources nodes, all sources nodes and
    # the resulting timeline needs to be located in the same reel.
    # The nodeId is the reel where the sources nodes will be created
    # and the timeline assembled.
    #
    parent_reel_node_handle = WireTapNodeHandle(server_handle, node_id)

    # Verify that the parent nodeId is effectively a reel
    #
    reel_type_str = WireTapStr("REEL") # IFFFS server specific type

    parent_sype_str = WireTapStr()
    if not parent_reel_node_handle.getNodeTypeStr(parent_sype_str):
        raise WireTapException("Unable to get the type of the parent node - %s." % \
                               parent_reel_node_handle.lastError())

    if not parent_sype_str == reel_type_str:
        raise WireTapException("Unable to assemble a timeline without a parent reel.")

    # Creation of 3 sources clips. Clip name, tape name and number of
    # frames are specified for each clip.
    #
    clip_node_handle_1 = WireTapNodeHandle()
    create_clip_node(
        parent_node_handle=parent_reel_node_handle,
        display_name="SOURCE_1",
        tape_name="TAPE1",
        num_frames=10,
        clip_node_handle=clip_node_handle_1
    )

    clip_node_handle_2 = WireTapNodeHandle()
    create_clip_node(
        parent_node_handle=parent_reel_node_handle,
        display_name="SOURCE_2",
        tape_name="TAPE2",
        num_frames=20,
        clip_node_handle=clip_node_handle_2
    )

    clip_node_handle_3 = WireTapNodeHandle()
    create_clip_node(
        parent_node_handle=parent_reel_node_handle,
        display_name="SOURCE_3",
        tape_name="TAPE2",
        num_frames=30,
        clip_node_handle=clip_node_handle_3
    )

    # Define an DMXEDL stream describing the timeline we want to
    # create. DMXEDL is an EDL type metadata stream specific to the
    # IFFFS Wiretap server.
    # This DMXEDL refers to the previously created sources clip.
    # The resulting timeline will have a cut between the first two
    # sources clips and a dissolve of 5 frames between the second and
    # third sources clips.
    # Note that the tape name indicated in the DMXEDL should match the
    # tape name of the sources clip previously created.
    metadata = (
        "TITLE: TIMELINE\n" \
        "FCM: NON-DROP FRAME\n\n" \
        "TITLE: ASSEMBLY RESOLUTION: " \
        "720:486:24:3:0.899998:1049776:BE:F1:29.969999    \n" \
        "FCM: NON-DROP FRAME\n\n" \
        "001  TAPE1   V     C        " \
        "00:00:00:00 00:00:00:10 00:00:00:10 00:00:00:20  \n" \
        "FROM CLIP NAME: SOURCE_1\n" \
        "DLEDL: START TC: 00:00:00:00\n\n" \
        "002  TAPE2   V     C        " \
        "00:00:00:00 00:00:00:13 00:00:00:20 00:00:01:03  \n" \
        "FROM CLIP NAME: SOURCE_2\n" \
        "DLEDL: START TC: 00:00:00:00\n\n" \
        "003  TAPE2   V     C        " \
        "00:00:00:13 00:00:00:13 00:00:01:03 00:00:01:03  \n" \
        "003  TAPE2   V     D    " \
        "005 00:00:00:03 00:00:00:20 00:00:01:03 00:00:02:00  \n" \
        "FROM CLIP NAME: SOURCE_2\n" \
        "TO CLIP NAME: SOURCE_3\n" \
        "DLEDL: START TC: 00:00:00:00\n" \
        "DLEDL: FOCUS_DESCR CENTERED\n")

    # Define a clip format.
    #
    clip_format = WireTapClipFormat(
        720, 486,   # width, height
        3 * 8,      # bits per pixel
        3, 	 # # channels
        29.97,	 # frame rate
        0.9,	 # pixel ratio
        WireTapClipFormat.ScanFormat.SCAN_FORMAT_FIELD_1_ODD,
        WireTapClipFormat.FORMAT_RGB()
    )

    # Create the timeline node which is a regular clip node
    #
    timeline_node_handle = WireTapNodeHandle()
    if not parent_reel_node_handle.createClipNode(
            "Timeline",             # display name
            clip_format,           # clip format
            "CLIP",                 # extended (server-specific) type
            timeline_node_handle):  # created node returned here
        raise WireTapException("Unable to create clip node: %s." % \
                               parent_reel_node_handle.lastError())

    print("Timeline node successfully created")

    # Apply the metadata stream to the created clip node.
    # The IFFFS timeline is assembled from the 3 sources clips.
    #
    if not timeline_node_handle.setMetaData("DMXEDL", metadata):
        raise WireTapException("Unable to apply the metadata stream: %s." %
                               timeline_node_handle.lastError())

    print("Timeline successfully assembled")


if __name__ == '__main__':
    main(sys.argv[1:])
