#!/usr/bin/env python

#
# Wiretap SDK tool for creating a generic node.
#

from __future__ import print_function

import getopt
import sys
from libwiretapPythonClientAPI import (
    WireTapClient,
    WireTapServerHandle,
    WireTapNodeHandle,
    WireTapMetaData
)

def printUsage():

    print(" Usage:")
    print("  -n <parent node ID>")
    print("  -t <server-specific node type string>")
    print("  -d <display name>")
    print("  [-h <host name or IP address> (default = localhost)]")

def main():

    # Check the command line options for correctness.
    try:
        opts, _ = getopt.getopt(sys.argv, "h:n:t:d:")
    except getopt.GetoptError:
        printUsage()
        sys.exit(2)

    # Declare variables (to be set from command line arguments).
    # For hostName, the default "localhost" will work if a Wiretap server
    # is running on your machine.
    # For nodeType, "NODE" is the type defined by the IFFFS Wiretap Server
    # for generic nodes.
    #
    hostName = "localhost"
    parentNodeId = ""
    nodeType = "NODE"
    displayName = "node"
    fileHandle = None
    streamName = WireTapMetaData.STREAM_XML()

    # Parse command line and set variables just declared.
    #
    for opt, arg in opts:
        if opt == '-h':
            hostName = arg
        elif opt == '-n':
            parentNodeId = arg
        if opt == "-t":
            nodeType = arg
        if opt == "-d":
            displayName = arg
        elif opt == '-s':
            streamName = arg
        elif opt == '-f':
            try:
                fileHandle = open(arg, 'r')
            except IOError as e:
                raise Exception("Cannot read meta data file '%s': %s" % (arg, e))

    if parentNodeId == "":
        print("Parent Node ID not specified.\n")
        printUsage()
        sys.exit(2)

    # Initialize the Wiretap Client API.
    #
    wiretapClient = WireTapClient()
    if not wiretapClient.init():
        raise Exception("Unable to initialize Wiretap client API")

    try:
        # Instantiate a server handle
        #
        server = WireTapServerHandle(hostName)

        # Instantiate the parent node
        #
        parentNode = WireTapNodeHandle(server, parentNodeId)
        node = WireTapNodeHandle()

        # Create the node.
        #
        if fileHandle:
            if not parentNode.createNode(displayName, nodeType, streamName,
                                         fileHandle.read(), node):
                raise Exception("Unable to create node: %s" % parentNode.lastError())
        else:
            if not parentNode.createNode(displayName, nodeType, node):
                raise Exception("Unable to create node: %s" % parentNode.lastError())

        print("Created node '%s'." % node.getNodeId().id())

    finally:
        # Must destroy WireTapServerHandle and WireTapNodeHandle before
        # uninitializing the Wiretap Client API.
        #
        node = None
        parentNode = None
        server = None

if __name__ == '__main__':
    main()
